// Complétez les TODO dans le fichier.
// Vous pouvez consulter les diapositives

import * as THREE from './build/three.module.js';
import Stats from './jsm/libs/stats.module.js';

import {
    ColladaLoader
}
    from './jsm/loaders/ColladaLoader.js';

import {
    OrbitControls
}
    from './jsm/controls/OrbitControls.js'

var THREEx = THREEx || {};

THREEx.KeyboardState = function (domElement) {
    this.domElement = domElement || document;
    // to store the current state
    this.keyCodes = {};
    this.modifiers = {};

    // create callback to bind/unbind keyboard events
    var _this = this;
    this._onKeyDown = function (event) {
        _this._onKeyChange(event)
    }
    this._onKeyUp = function (event) {
        _this._onKeyChange(event)
    }

    // bind keyEvents
    this.domElement.addEventListener("keydown", this._onKeyDown, false);
    this.domElement.addEventListener("keyup", this._onKeyUp, false);

    // create callback to bind/unbind window blur event
    this._onBlur = function () {
        for (var prop in _this.keyCodes)
            _this.keyCodes[prop] = false;
        for (var prop in _this.modifiers)
            _this.modifiers[prop] = false;
    }

    // bind window blur
    window.addEventListener("blur", this._onBlur, false);
}

THREEx.KeyboardState.prototype.destroy = function () {
    // unbind keyEvents
    this.domElement.removeEventListener("keydown", this._onKeyDown, false);
    this.domElement.removeEventListener("keyup", this._onKeyUp, false);

    // unbind window blur event
    window.removeEventListener("blur", this._onBlur, false);
}

THREEx.KeyboardState.MODIFIERS = ['shift', 'ctrl', 'alt', 'meta'];
THREEx.KeyboardState.ALIAS = {
    'left': 37,
    'up': 38,
    'right': 39,
    'down': 40,
    'space': 32,
    'pageup': 33,
    'pagedown': 34,
    'tab': 9,
    'escape': 27
};

/**
 * to process the keyboard dom event
 */
THREEx.KeyboardState.prototype._onKeyChange = function (event) {
    // log to debug
    //console.log("onKeyChange", event, event.keyCode, event.shiftKey, event.ctrlKey, event.altKey, event.metaKey)

    // update this.keyCodes
    var keyCode = event.keyCode
    var pressed = event.type === 'keydown' ? true : false
    this.keyCodes[keyCode] = pressed
    // update this.modifiers
    this.modifiers['shift'] = event.shiftKey
    this.modifiers['ctrl'] = event.ctrlKey
    this.modifiers['alt'] = event.altKey
    this.modifiers['meta'] = event.metaKey
}

/**
 * query keyboard state to know if a key is pressed of not
 *
 * @param {String} keyDesc the description of the key. format : modifiers+key e.g shift+A
 * @returns {Boolean} true if the key is pressed, false otherwise
 */
THREEx.KeyboardState.prototype.pressed = function (keyDesc) {
    var keys = keyDesc.split("+");
    for (var i = 0; i < keys.length; i++) {
        var key = keys[i]
        var pressed = false
        if (THREEx.KeyboardState.MODIFIERS.indexOf(key) !== -1) {
            pressed = this.modifiers[key];
        } else if (Object.keys(THREEx.KeyboardState.ALIAS).indexOf(key) != -1) {
            pressed = this.keyCodes[THREEx.KeyboardState.ALIAS[key]];
        } else {
            pressed = this.keyCodes[key.toUpperCase().charCodeAt(0)]
        }
        if (!pressed)
            return false;
    };
    return true;
}

/**
 * return true if an event match a keyDesc
 * @param  {KeyboardEvent} event   keyboard event
 * @param  {String} keyDesc string description of the key
 * @return {Boolean}         true if the event match keyDesc, false otherwise
 */
THREEx.KeyboardState.prototype.eventMatches = function (event, keyDesc) {
    var aliases = THREEx.KeyboardState.ALIAS
    var aliasKeys = Object.keys(aliases)
    var keys = keyDesc.split("+")
    // log to debug
    // console.log("eventMatches", event, event.keyCode, event.shiftKey, event.ctrlKey, event.altKey, event.metaKey)
    for (var i = 0; i < keys.length; i++) {
        var key = keys[i];
        var pressed = false;
        if (key === 'shift') {
            pressed = (event.shiftKey ? true : false)
        } else if (key === 'ctrl') {
            pressed = (event.ctrlKey ? true : false)
        } else if (key === 'alt') {
            pressed = (event.altKey ? true : false)
        } else if (key === 'meta') {
            pressed = (event.metaKey ? true : false)
        } else if (aliasKeys.indexOf(key) !== -1) {
            pressed = (event.keyCode === aliases[key] ? true : false);
        } else if (event.keyCode === key.toUpperCase().charCodeAt(0)) {
            pressed = true;
        }
        if (!pressed)
            return false;
    }
    return true;
}

let container, stats, clock, controls;
let lights, camera, scene, renderer, humanGeometry, humanMaterial, humanMesh, robot;
let skinWeight, skinIndices, boneArray, realBones, boneDict;

THREE.Cache.enabled = true;


THREE.Object3D.prototype.setMatrix = function (a) {
    this.matrix = a;
    this.matrix.decompose(this.position, this.quaternion, this.scale);
};


class Robot {
    constructor(h) {
        this.spineLength = 0.65305;
        this.chestLength = 0.46487;
        this.neckLength = 0.24523
        this.headLength = 0.39284;

        this.armLength = 0.72111;
        this.forearmLength = 0.61242;
        this.legLength = 1.16245;
        this.shinLength = 1.03432;

        this.armLeftRotation = realBones[4].rotation;
        this.forearmLeftRotation = realBones[5].rotation;
        this.armRightRotation = realBones[6].rotation;
        this.forearmRightRotation = realBones[7].rotation;

        this.legLeftRotation = realBones[8].rotation;
        this.shinLeftRotation = realBones[9].rotation;
        this.legRightRotation = realBones[10].rotation;
        this.shinRightRotation = realBones[11].rotation;

        this.spineTranslation = realBones[0].position;
        this.chestTranslation = realBones[1].position;
        this.neckTranslation = realBones[2].position;
        this.headTranslation = realBones[3].position;
        this.armLeftTranslation = realBones[4].position;

        this.forearmLeftTranslation = realBones[5].position;
        this.armRightTranslation = realBones[6].position;
        this.forearmRightTranslation = realBones[7].position;

        this.legLeftTranslation = realBones[8].position;
        this.shinLeftTranslation = realBones[9].position;
        this.legRightTranslation = realBones[10].position;
        this.shinRightTranslation = realBones[11].position;

        this.bodyWidth = 0.2;
        this.bodyDepth = 0.2;


        this.neckRadius = 0.1;

        this.headRadius = 0.32;


        this.legRadius = 0.10;
        this.thighRadius = 0.1;
        this.footDepth = 0.4;
        this.footWidth = 0.25;
        this.footHeight = 0.1;

        this.armRadius = 0.10;

        this.handRadius = 0.1;

        // Material
        this.material = new THREE.MeshNormalMaterial();
        this.human = h;
        // Initial pose
        this.initialize()
    }

    initialize() {
        // X: Rouge
        // Y: Vert
        // Z: Bleu
        if (!this.hasOwnProperty("axes"))
            this.axes = new THREE.AxesHelper(2);

        // Spine geometry
        var spineGeometry = new THREE.CylinderGeometry(0.5 * this.bodyWidth / 2, this.bodyWidth / 2, this.spineLength, 64);
        if (!this.hasOwnProperty("spine"))
            this.spine = new THREE.Mesh(spineGeometry, this.material);

        // Chest geometry
        var chestGeometry = new THREE.CylinderGeometry(0.5 * this.bodyWidth / 2, this.bodyWidth / 2, this.chestLength, 64);
        if (!this.hasOwnProperty("chest"))
            this.chest = new THREE.Mesh(chestGeometry, this.material);

        // Neck geometry
        var neckGeometry = new THREE.CylinderGeometry(0.5 * this.neckRadius, this.neckRadius, this.neckLength, 64);
        if (!this.hasOwnProperty("neck"))
            this.neck = new THREE.Mesh(neckGeometry, this.material);

        // Head geometry
        var headGeometry = new THREE.SphereGeometry(this.headLength / 2, 64, 10);
        if (!this.hasOwnProperty("head"))
            this.head = new THREE.Mesh(headGeometry, this.material);

        // Left arm geometry
        var arm_LGeometry = new THREE.CylinderGeometry(0.5 * this.armRadius, this.armRadius, this.armLength, 64);
        if (!this.hasOwnProperty("arm_L"))
            this.arm_L = new THREE.Mesh(arm_LGeometry, this.material);

        // Right arm geometry
        var arm_RGeometry = new THREE.CylinderGeometry(0.5 * this.armRadius, this.armRadius, this.armLength, 64);
        if (!this.hasOwnProperty("arm_R"))
            this.arm_R = new THREE.Mesh(arm_RGeometry, this.material);

        // Left forearm geometry
        var forearm_LGeometry = new THREE.CylinderGeometry(0.5 * this.armRadius, this.armRadius, this.forearmLength, 64);
        if (!this.hasOwnProperty("forearm_L"))
            this.forearm_L = new THREE.Mesh(forearm_LGeometry, this.material);

        // Right forearm geometry
        var forearm_RGeometry = new THREE.CylinderGeometry(0.5 * this.armRadius, this.armRadius, this.forearmLength, 64);
        if (!this.hasOwnProperty("forearm_R"))
            this.forearm_R = new THREE.Mesh(forearm_RGeometry, this.material);

        //Left hand geometry
        var hand_LGeometry = new THREE.SphereGeometry(this.handRadius, 64, 10);
        if (!this.hasOwnProperty("hand_L"))
            this.hand_L = new THREE.Mesh(hand_LGeometry, this.material);

        // Right hand geometry
        var hand_RGeometry = new THREE.SphereGeometry(this.handRadius, 64, 10);
        if (!this.hasOwnProperty("hand_R"))
            this.hand_R = new THREE.Mesh(hand_RGeometry, this.material);

        // Left leg geometry
        var leg_LGeometry = new THREE.CylinderGeometry(0.5 * this.thighRadius, this.thighRadius, this.legLength, 64);
        if (!this.hasOwnProperty("leg_L"))
            this.leg_L = new THREE.Mesh(leg_LGeometry, this.material);

        // Right leg geometry
        var leg_RGeometry = new THREE.CylinderGeometry(0.5 * this.thighRadius, this.thighRadius, this.legLength, 64);
        if (!this.hasOwnProperty("leg_R"))
            this.leg_R = new THREE.Mesh(leg_RGeometry, this.material);

        // Left shin geometry
        var shin_LGeometry = new THREE.CylinderGeometry(0.5 * this.legRadius, this.legRadius, this.shinLength, 64);
        if (!this.hasOwnProperty("shin_L"))
            this.shin_L = new THREE.Mesh(shin_LGeometry, this.material);

        // Right shin geometry
        var shin_RGeometry = new THREE.CylinderGeometry(0.5 * this.legRadius, this.legRadius, this.shinLength, 64);
        if (!this.hasOwnProperty("shin_R"))
            this.shin_R = new THREE.Mesh(shin_RGeometry, this.material);

        // Left foot matrix
        var foot_LGeometry = new THREE.BoxGeometry(this.footWidth, this.footHeight, this.footDepth, 1, 1, 1);
        if (!this.hasOwnProperty("foot_L"))
            this.foot_L = new THREE.Mesh(foot_LGeometry, this.material);

        // Right foot geometry
        var foot_RGeometry = new THREE.BoxGeometry(this.footWidth, this.footHeight, this.footDepth, 1, 1, 1);
        if (!this.hasOwnProperty("foot_R"))
            this.foot_R = new THREE.Mesh(foot_RGeometry, this.material);

        // Spine matrix
        this.localSpineMatrix = translation(0.0, this.spineTranslation.y + this.spineLength / 2, 0.0);

        // Chest matrix
        this.localChestMatrix = translation(0.0, this.chestTranslation.y - this.spineLength / 2 + this.chestLength / 2, 0.0);
        var chestMatrix = matMul(this.localSpineMatrix, this.localChestMatrix);

        // Neck matrix
        this.localNeckMatrix = translation(0.0, this.neckTranslation.y - this.chestLength / 2 + this.neckLength / 2, 0.0);
        var neckMatrix = matMul(chestMatrix, this.localNeckMatrix);

        // Head matrix
        this.localHeadMatrix = translation(0.0, this.headTranslation.y - this.neckLength / 2 + this.headLength / 2, 0.0);
        var headMatrix = matMul(neckMatrix, this.localHeadMatrix);



        // Jambe Gauche
        // Left leg matrix
        this.localLegLMatrix = matMul(translation(this.legLeftTranslation.x, this.legLeftTranslation.y, this.legLeftTranslation.z),
            matMul(translation(0, -1 * this.spineLength / 2, 0),
                matMul(rotX(this.legLeftRotation.x),
                    matMul(rotY(this.legLeftRotation.y),
                        matMul(rotZ(this.legLeftRotation.z), translation(0, this.legLength / 2, 0))))));
        var leg_LMatrix = matMul(this.localSpineMatrix, this.localLegLMatrix);

        // Left shin matrix
        this.localShinLMatrix = matMul(translation(this.shinLeftTranslation.x, this.shinLeftTranslation.y, this.shinLeftTranslation.z),
            matMul(translation(0, -1 * this.legLength / 2, 0),
                matMul(rotX(this.shinLeftRotation.x),
                    matMul(rotY(this.shinLeftRotation.y),
                        matMul(rotZ(this.shinLeftRotation.z), translation(0, this.shinLength / 2, 0))))));
        var shin_LMatrix = matMul(leg_LMatrix, this.localShinLMatrix);


        // Left foot matrix
        this.localFootLMatrix = translation(0, this.footHeight + this.shinLength / 1.7, this.footDepth / 3);
        var foot_LMatrix = matMul(shin_LMatrix, this.localFootLMatrix);


        // TODO 1: Complétez le squelette
        // Remplacez les translations et "new THREE.Matrix4()" avec votre code.
        // Pour la jambe droite, inspirez-vous de la jambe gauche! Vous pouvez vous référez aux diapositives.
        // N'hésitez pas de demander de l'aide.

        // Right leg matrix
        this.localLegRMatrix = translation(2, 0, 0); 
        var leg_RMatrix = new THREE.Matrix4();

        // Right shin matrix
        this.localShinRMatrix = translation(2, 0, 0); 
        var shin_RMatrix = new THREE.Matrix4();

        // Right foot matrix
        this.localFootRMatrix = translation(2, 0, 0); 
        var foot_RMatrix = new THREE.Matrix4(); 

        // LEFT ARM
        // Left arm matrix
        this.localArmLMatrix = translation(2, 0, 0); 
        var arm_LMatrix = new THREE.Matrix4(); 

        // Left forearm matrix
        this.localForearmLMatrix = translation(2, 0, 0); 
        var forearm_LMatrix = new THREE.Matrix4(); 

        // RIGHT ARM
        // Right arm matrix
        this.localArmRMatrix = translation(2, 0, 0); 
        var arm_RMatrix = new THREE.Matrix4();  

        // Right forearm matrix
        this.localForearmRMatrix = translation(2, 0, 0); 
        var forearm_RMatrix = new THREE.Matrix4();

        // FIN de TODO 1
        // Left hand matrix
        this.localHandLMatrix = translation(0, this.handRadius + this.forearmLength / 2, 0);
        var hand_LMatrix = matMul(forearm_LMatrix, this.localHandLMatrix);


        // Right hand matrix
        this.localHandRMatrix = translation(0, this.handRadius + this.forearmLength / 2, 0);
        var hand_RMatrix = matMul(forearm_RMatrix, this.localHandRMatrix);

        // Apply transformations
        this.spine.setMatrix(this.localSpineMatrix);
        if (scene.getObjectById(this.spine.id) === undefined)
            scene.add(this.spine);

        this.chest.setMatrix(chestMatrix);
        if (scene.getObjectById(this.chest.id) === undefined)
            scene.add(this.chest);

        this.neck.setMatrix(neckMatrix);
        if (scene.getObjectById(this.neck.id) === undefined)
            scene.add(this.neck);

        this.head.setMatrix(headMatrix);
        if (scene.getObjectById(this.head.id) === undefined)
            scene.add(this.head);

        this.arm_L.setMatrix(arm_LMatrix);
        if (scene.getObjectById(this.arm_L.id) === undefined)
            scene.add(this.arm_L);

        this.arm_R.setMatrix(arm_RMatrix);
        if (scene.getObjectById(this.arm_R.id) === undefined)
            scene.add(this.arm_R);

        this.forearm_L.setMatrix(forearm_LMatrix);
        if (scene.getObjectById(this.forearm_L.id) === undefined)
            scene.add(this.forearm_L);

        this.forearm_R.setMatrix(forearm_RMatrix);
        if (scene.getObjectById(this.forearm_R.id) === undefined)
            scene.add(this.forearm_R);

        this.hand_L.setMatrix(hand_LMatrix);
        if (scene.getObjectById(this.hand_L.id) === undefined)
            scene.add(this.hand_L);

        this.hand_R.setMatrix(hand_RMatrix);
        if (scene.getObjectById(this.hand_R.id) === undefined)
            scene.add(this.hand_R);

        this.leg_L.setMatrix(leg_LMatrix);
        if (scene.getObjectById(this.leg_L.id) === undefined)
            scene.add(this.leg_L);

        this.leg_R.setMatrix(leg_RMatrix);
        if (scene.getObjectById(this.leg_R.id) === undefined)
            scene.add(this.leg_R);

        this.shin_L.setMatrix(shin_LMatrix);
        if (scene.getObjectById(this.shin_L.id) === undefined)
            scene.add(this.shin_L);

        this.shin_R.setMatrix(shin_RMatrix);
        if (scene.getObjectById(this.shin_R.id) === undefined)
            scene.add(this.shin_R);

        this.foot_L.setMatrix(foot_LMatrix);
        if (scene.getObjectById(this.foot_L.id) === undefined)
            scene.add(this.foot_L);

        this.foot_R.setMatrix(foot_RMatrix);
        if (scene.getObjectById(this.foot_R.id) === undefined)
            scene.add(this.foot_R);

        if (scene.getObjectById(this.axes) === undefined)
            scene.add(this.axes)

    }

    hideRobot() {
        this.spine.visible = false;
        this.chest.visible = false;
        this.neck.visible = false;
        this.head.visible = false;
        this.arm_R.visible = false;
        this.arm_L.visible = false;
        this.forearm_L.visible = false;
        this.forearm_R.visible = false;
        this.hand_L.visible = false;
        this.hand_R.visible = false;
        this.leg_L.visible = false;
        this.leg_R.visible = false;
        this.shin_L.visible = false;
        this.shin_R.visible = false;
        this.foot_L.visible = false;
        this.foot_R.visible = false;
    }

    hideHuman() {
        this.human.visible = false;
    }

    showRobot() {
        this.spine.visible = true;
        this.chest.visible = true;
        this.neck.visible = true;
        this.head.visible = true;
        this.arm_R.visible = true;
        this.arm_L.visible = true;
        this.forearm_L.visible = true;
        this.forearm_R.visible = true;
        this.hand_L.visible = true;
        this.hand_R.visible = true;
        this.leg_L.visible = true;
        this.leg_R.visible = true;
        this.shin_L.visible = true;
        this.shin_R.visible = true;
        this.foot_L.visible = true;
        this.foot_R.visible = true;
    }

    showHuman() {
        this.human.visible = true;
    }

    update_for_skinning() {
        //Envoi des matrices au vertex shader pour le mesh
        buildSkeleton();
        boneDict["Spine"].setMatrix(matMul(matMul(this.spine.matrix, translation(0, -1 * this.spineLength / 2, 0)), getInverse(boneDict["Spine"].matrixWorld)));
        boneDict["Chest"].setMatrix(matMul(matMul(this.chest.matrix, translation(0, -1 * this.chestLength / 2, 0)), getInverse(boneDict["Chest"].matrixWorld)));
        boneDict["Neck"].setMatrix(matMul(matMul(this.neck.matrix, translation(0, -1 * this.neckLength / 2, 0)), getInverse(boneDict["Neck"].matrixWorld)));
        boneDict["Head"].setMatrix(matMul(matMul(this.head.matrix, translation(0, -1 * this.headLength / 2, 0)), getInverse(boneDict["Head"].matrixWorld)));
        boneDict["Arm_L"].setMatrix(matMul(matMul(this.arm_L.matrix, translation(0, -1 * this.armLength / 2, 0)), getInverse(boneDict["Arm_L"].matrixWorld)));
        boneDict["Arm_R"].setMatrix(matMul(matMul(this.arm_R.matrix, translation(0, -1 * this.armLength / 2, 0)), getInverse(boneDict["Arm_R"].matrixWorld)));
        boneDict["Forearm_L"].setMatrix(matMul(matMul(this.forearm_L.matrix, translation(0, -1 * this.forearmLength / 2, 0)), getInverse(boneDict["Forearm_L"].matrixWorld)));
        boneDict["Forearm_R"].setMatrix(matMul(matMul(this.forearm_R.matrix, translation(0, -1 * this.forearmLength / 2, 0)), getInverse(boneDict["Forearm_R"].matrixWorld)));
        boneDict["Leg_L"].setMatrix(matMul(matMul(this.leg_L.matrix, translation(0, -1 * this.legLength / 2, 0)), getInverse(boneDict["Leg_L"].matrixWorld)));
        boneDict["Leg_R"].setMatrix(matMul(matMul(this.leg_R.matrix, translation(0, -1 * this.legLength / 2, 0)), getInverse(boneDict["Leg_R"].matrixWorld)));
        boneDict["Shin_L"].setMatrix(matMul(matMul(this.shin_L.matrix, translation(0, -1 * this.shinLength / 2, 0)), getInverse(boneDict["Shin_L"].matrixWorld)));
        boneDict["Shin_R"].setMatrix(matMul(matMul(this.shin_R.matrix, translation(0, -1 * this.shinLength / 2, 0)), getInverse(boneDict["Shin_R"].matrixWorld)));
        buildShaderBoneMatrix();
    }


    pose1() {
        var offSet = 0.12;

        this.spine.setMatrix(matMul(this.localSpineMatrix, translation(0, -1 * (this.footHeight + this.shinLength) + offSet, 0)));
        this.chest.setMatrix(matMul(this.spine.matrix, this.localChestMatrix));
        this.neck.setMatrix(matMul(this.chest.matrix, this.localNeckMatrix));
        this.head.setMatrix(matMul(this.neck.matrix, this.localHeadMatrix));

        this.arm_L.setMatrix(matMul(this.chest.matrix,
            matMul(this.localArmLMatrix,
                matMul(translation(0, -1 * this.armLength / 2, 0),
                    matMul(rotZ(Math.PI / 10),
                        translation(0, this.armLength / 2, 0))))));

        this.arm_R.setMatrix(matMul(this.chest.matrix,
            matMul(this.localArmRMatrix,
                matMul(translation(0, -1 * this.armLength / 2, 0),
                    matMul(rotX(Math.PI / 4),
                        translation(0, this.armLength / 2, 0))))));

        this.forearm_L.setMatrix(matMul(this.arm_L.matrix,
            matMul(this.localForearmLMatrix,
                matMul(translation(0, -1 * this.forearmLength / 2, 0),
                    matMul(rotZ(-1 * Math.PI / 1.7),
                        translation(0, this.forearmLength / 2, 0))))));

        this.forearm_R.setMatrix(matMul(this.arm_R.matrix,
            matMul(this.localForearmRMatrix,
                matMul(translation(0, -1 * this.forearmLength / 2, 0),
                    matMul(rotX(Math.PI / 4),
                        matMul(rotY(Math.PI / 3),
                            matMul(rotZ(Math.PI / 2),
                                translation(0, this.forearmLength / 2, 0))))))));

        this.hand_L.setMatrix(matMul(this.forearm_L.matrix, this.localHandLMatrix));
        this.hand_R.setMatrix(matMul(this.forearm_R.matrix, this.localHandRMatrix));

        this.leg_L.setMatrix(matMul(this.spine.matrix, this.localLegLMatrix));

        this.leg_R.setMatrix(matMul(this.spine.matrix,
            matMul(this.localLegRMatrix,
                matMul(translation(0, -1 * this.legLength / 2, 0),
                    matMul(rotX(Math.PI / 2.2),
                        translation(0, this.legLength / 2, 0))))));

        this.shin_L.setMatrix(matMul(this.leg_L.matrix,
            matMul(this.localShinLMatrix,
                matMul(translation(0, -1 * this.shinLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 2),
                        translation(0, this.shinLength / 2, 0))))));

        this.shin_R.setMatrix(matMul(this.leg_R.matrix,
            matMul(this.localShinRMatrix,
                matMul(translation(0, -1 * this.shinLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 2.2),
                        translation(0, this.shinLength / 2, 0))))));

        this.foot_L.setMatrix(matMul(this.shin_L.matrix, this.localFootLMatrix));
        this.foot_R.setMatrix(matMul(this.shin_R.matrix, this.localFootRMatrix));
        this.update_for_skinning();
    }

    pose2() {
        var jumpHeight = 2;

        this.spine.setMatrix(matMul(this.localSpineMatrix, translation(0, jumpHeight, 0)));
        this.chest.setMatrix(matMul(this.spine.matrix, this.localChestMatrix));
        this.neck.setMatrix(matMul(this.chest.matrix, this.localNeckMatrix));
        this.head.setMatrix(matMul(this.neck.matrix, this.localHeadMatrix));

        this.arm_L.setMatrix(matMul(this.chest.matrix,
            matMul(this.localArmLMatrix,
                matMul(translation(0, -1 * this.armLength / 2, 0),
                    matMul(rotY(-1 * Math.PI / 8),
                        matMul(rotX(Math.PI / 2),
                            translation(0, this.armLength / 2, 0)))))));

        this.arm_R.setMatrix(matMul(this.chest.matrix,
            matMul(this.localArmRMatrix,
                matMul(translation(0, -1 * this.armLength / 2, 0),
                    matMul(rotY(Math.PI / 6),
                        matMul(rotZ(-1 * Math.PI / 4),
                            translation(0, this.armLength / 2, 0)))))));

        this.forearm_L.setMatrix(matMul(this.arm_L.matrix, this.localForearmLMatrix));

        this.forearm_R.setMatrix(matMul(this.arm_R.matrix,
            matMul(this.localForearmRMatrix,
                matMul(translation(0, -1 * this.forearmLength / 2, 0),
                    matMul(rotY(-1 * Math.PI / 8),
                        matMul(rotZ(-1 * Math.PI / 2),
                            matMul(rotY(Math.PI / 2),
                                translation(0, this.forearmLength / 2, 0))))))));

        this.hand_L.setMatrix(matMul(this.forearm_L.matrix, this.localHandLMatrix));
        this.hand_R.setMatrix(matMul(this.forearm_R.matrix, this.localHandRMatrix));

        this.leg_L.setMatrix(matMul(this.spine.matrix,
            matMul(this.localLegLMatrix,
                matMul(translation(0, -1 * this.legLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 5),
                        translation(0, this.legLength / 2, 0))))));

        this.leg_R.setMatrix(matMul(this.spine.matrix,
            matMul(this.localLegRMatrix,
                matMul(translation(0, -1 * this.legLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 5),
                        translation(0, this.legLength / 2, 0))))));

        this.shin_L.setMatrix(matMul(this.leg_L.matrix,
            matMul(this.localShinLMatrix,
                matMul(translation(0, -1 * this.shinLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 1.8),
                        translation(0, this.shinLength / 2, 0))))));

        this.shin_R.setMatrix(matMul(this.leg_R.matrix,
            matMul(this.localShinRMatrix,
                matMul(translation(0, -1 * this.shinLength / 2, 0),
                    matMul(rotX(-1 * Math.PI / 1.8),
                        translation(0, this.shinLength / 2, 0))))));

        this.foot_L.setMatrix(matMul(this.shin_L.matrix, this.localFootLMatrix));
        this.foot_R.setMatrix(matMul(this.shin_R.matrix, this.localFootRMatrix));

        //Envoi des matrices au vertex shader pour le mesh
        this.update_for_skinning();
    }

    pose_rest() {
        this.spine.setMatrix(this.localSpineMatrix);
        this.chest.setMatrix(matMul(this.spine.matrix, this.localChestMatrix));
        this.neck.setMatrix(matMul(this.chest.matrix, this.localNeckMatrix));
        this.head.setMatrix(matMul(this.neck.matrix, this.localHeadMatrix));

        this.arm_L.setMatrix(matMul(this.chest.matrix, this.localArmLMatrix));
        this.arm_R.setMatrix(matMul(this.chest.matrix, this.localArmRMatrix));

        this.forearm_L.setMatrix(matMul(this.arm_L.matrix, this.localForearmLMatrix));
        this.forearm_R.setMatrix(matMul(this.arm_R.matrix, this.localForearmRMatrix));

        this.hand_L.setMatrix(matMul(this.forearm_L.matrix, this.localHandLMatrix));
        this.hand_R.setMatrix(matMul(this.forearm_R.matrix, this.localHandRMatrix));

        this.leg_L.setMatrix(matMul(this.spine.matrix, this.localLegLMatrix));
        this.leg_R.setMatrix(matMul(this.spine.matrix, this.localLegRMatrix))

        this.shin_L.setMatrix(matMul(this.leg_L.matrix, this.localShinLMatrix));
        this.shin_R.setMatrix(matMul(this.leg_R.matrix, this.localShinRMatrix));

        this.foot_L.setMatrix(matMul(this.shin_L.matrix, this.localFootLMatrix));
        this.foot_R.setMatrix(matMul(this.shin_R.matrix, this.localFootRMatrix));

        this.update_for_skinning();
    }

    runAnimation(t) {
        var jumpHeight = 0.25;
        var speed = 7;
        var armsAngle = Math.PI / 4;
        var forearmsAngle = Math.PI / 2;
        var legsAngle = Math.PI / 4;
        var shinAngle = Math.PI / 2;

        this.spine.setMatrix(matMul(this.localSpineMatrix, translation(0, jumpHeight * (sin(speed * t) + 1.0) / 2.0, 0)));

        this.chest.setMatrix(matMul(this.spine.matrix,
            matMul(this.localChestMatrix,
                matMul(translation(0, -1 * this.chestLength / 2, 0),
                    matMul(rotY(-1 * Math.PI / 5 * sin(speed * t)),
                        translation(0, this.chestLength / 2, 0))))));

        this.neck.setMatrix(matMul(this.chest.matrix, this.localNeckMatrix));
        this.head.setMatrix(matMul(this.neck.matrix, this.localHeadMatrix));

        this.arm_L.setMatrix(matMul(this.chest.matrix,
            matMul(this.localArmLMatrix,
                matMul(translation(0, -1 * this.armLength / 2, 0),
                    matMul(rotX(armsAngle * sin(speed * t)),
                        translation(0, this.armLength / 2, 0))))));
        this.forearm_L.setMatrix(matMul(this.arm_L.matrix,
            matMul(this.localForearmLMatrix,
                matMul(translation(0, -1 * this.forearmLength / 2, 0),
                    matMul(rotX(forearmsAngle * (sin(speed * t) + 1.0) / 2.0),
                        translation(0, this.forearmLength / 2, 0))))));

        this.hand_L.setMatrix(matMul(this.forearm_L.matrix, this.localHandLMatrix));


        // TODO 2: Complétez l'animation du squelettes en utilisant la variable "t" et
        // des fonctions trigonométriques

        this.arm_R.setMatrix(matMul(this.chest.matrix, this.localArmRMatrix));
        this.forearm_R.setMatrix(matMul(this.arm_R.matrix, this.localForearmRMatrix));
        this.hand_R.setMatrix(matMul(this.forearm_R.matrix, this.localHandRMatrix));

        this.leg_L.setMatrix(matMul(this.spine.matrix, this.localLegLMatrix));
        this.shin_L.setMatrix(matMul(this.leg_L.matrix, this.localShinLMatrix));
        this.foot_L.setMatrix(matMul(this.shin_L.matrix, this.localFootLMatrix));

        this.leg_R.setMatrix(matMul(this.spine.matrix, this.localLegRMatrix))
        this.shin_R.setMatrix(matMul(this.leg_R.matrix, this.localShinRMatrix));
        this.foot_R.setMatrix(matMul(this.shin_R.matrix, this.localFootRMatrix));

        this.update_for_skinning();
    }


    customAnimation(t) {

        // TODO 3: Amusez-vous en faisant votre propre animation!
        // Faîtes quelque chose d'intéressant!

        this.spine.setMatrix(this.localSpineMatrix);
        this.chest.setMatrix(matMul(this.spine.matrix, this.localChestMatrix));
        this.neck.setMatrix(matMul(this.chest.matrix, this.localNeckMatrix));
        this.head.setMatrix(matMul(this.neck.matrix, this.localHeadMatrix));

        this.arm_L.setMatrix(matMul(this.chest.matrix, this.localArmLMatrix));
        this.arm_R.setMatrix(matMul(this.chest.matrix, this.localArmRMatrix));

        this.forearm_L.setMatrix(matMul(this.arm_L.matrix, this.localForearmLMatrix));
        this.forearm_R.setMatrix(matMul(this.arm_R.matrix, this.localForearmRMatrix));

        this.hand_L.setMatrix(matMul(this.forearm_L.matrix, this.localHandLMatrix));
        this.hand_R.setMatrix(matMul(this.forearm_R.matrix, this.localHandRMatrix));

        this.leg_L.setMatrix(matMul(this.spine.matrix, this.localLegLMatrix));
        this.leg_R.setMatrix(matMul(this.spine.matrix, this.localLegRMatrix))

        this.shin_L.setMatrix(matMul(this.leg_L.matrix, this.localShinLMatrix));
        this.shin_R.setMatrix(matMul(this.leg_R.matrix, this.localShinRMatrix));

        this.foot_L.setMatrix(matMul(this.shin_L.matrix, this.localFootLMatrix));
        this.foot_R.setMatrix(matMul(this.shin_R.matrix, this.localFootRMatrix));

        this.update_for_skinning();
    }
}

var keyboard = new THREEx.KeyboardState();
var channel = 'p';
var pi = Math.PI;

function init() {

    container = document.getElementById('container');

    camera = new THREE.PerspectiveCamera(45, window.innerWidth / window.innerHeight, 0.1, 2000);
    camera.position.set(8, 10, 8);
    camera.lookAt(0, 3, 0);

    scene = new THREE.Scene();
    scene.add(camera);

    controls = new OrbitControls(camera, container);
    controls.damping = 0.2;

    clock = new THREE.Clock();

    boneDict = {}

    boneArray = new Float32Array(12 * 16);

    humanMaterial = new THREE.ShaderMaterial({
        uniforms: {
            bones: {
                value: boneArray
            }
        },
        transparent: true,
    });

    const shaderLoader = new THREE.FileLoader();
    shaderLoader.load('glsl/human.vs.glsl',
        function (data) {
            humanMaterial.vertexShader = data;
        })
    shaderLoader.load('glsl/human.fs.glsl',
        function (data) {
            humanMaterial.fragmentShader = data;
        })

    // loading manager

    const loadingManager = new THREE.LoadingManager(function () {
        scene.add(humanMesh);
    });

    // collada
    humanGeometry = new THREE.BufferGeometry();
    const loader = new ColladaLoader(loadingManager);
    loader.load('./model/human.dae', function (collada) {
        skinIndices = collada.library.geometries['human-mesh'].build.triangles.data.attributes.skinIndex.array;
        skinWeight = collada.library.geometries['human-mesh'].build.triangles.data.attributes.skinWeight.array;
        realBones = collada.library.nodes.human.build.skeleton.bones;

        buildSkeleton();
        buildShaderBoneMatrix();
        humanGeometry.setAttribute('position', new THREE.BufferAttribute(collada.library.geometries['human-mesh'].build.triangles.data.attributes.position.array, 3));
        humanGeometry.setAttribute('skinWeight', new THREE.BufferAttribute(skinWeight, 4));
        humanGeometry.setAttribute('skinIndex', new THREE.BufferAttribute(skinIndices, 4));
        humanGeometry.setAttribute('normal', new THREE.BufferAttribute(collada.library.geometries['human-mesh'].build.triangles.data.attributes.normal.array, 3));

        humanMesh = new THREE.Mesh(humanGeometry, humanMaterial);
        robot = new Robot(humanMesh);
        robot.hideHuman();

    });

    //

    const ambientLight = new THREE.AmbientLight(0xcccccc, 0.4);
    scene.add(ambientLight);

    const directionalLight = new THREE.DirectionalLight(0xffffff, 0.8);
    directionalLight.position.set(1, 1, 0).normalize();
    scene.add(directionalLight);

    //

    renderer = new THREE.WebGLRenderer();
    renderer.setPixelRatio(window.devicePixelRatio);
    renderer.setSize(window.innerWidth, window.innerHeight);
    container.appendChild(renderer.domElement);

    //

    stats = new Stats();
    container.appendChild(stats.dom);

    //

    window.addEventListener('resize', onWindowResize);
    lights = [];
    lights[0] = new THREE.PointLight(0xffffff, 1, 0);
    lights[1] = new THREE.PointLight(0xffffff, 1, 0);
    lights[2] = new THREE.PointLight(0xffffff, 1, 0);

    lights[0].position.set(0, 200, 0);
    lights[1].position.set(100, 200, 100);
    lights[2].position.set(- 100, - 200, - 100);

    scene.add(lights[0]);
    scene.add(lights[1]);
    scene.add(lights[2]);

    var floorTexture = new THREE.ImageUtils.loadTexture('textures/hardwood2_diffuse.jpg');
    floorTexture.wrapS = floorTexture.wrapT = THREE.RepeatWrapping;
    floorTexture.repeat.set(4, 4);

    var floorMaterial = new THREE.MeshBasicMaterial({
        map: floorTexture,
        side: THREE.DoubleSide
    });
    var floorGeometry = new THREE.PlaneBufferGeometry(30, 30);
    var floor = new THREE.Mesh(floorGeometry, floorMaterial);
    floor.rotation.x = Math.PI / 2;
    floor.position.y -= 2.65;
    scene.add(floor);


}

function buildSkeleton() {
    boneDict["Spine"] = new THREE.Bone();
    boneDict["Chest"] = new THREE.Bone();
    boneDict["Neck"] = new THREE.Bone();
    boneDict["Head"] = new THREE.Bone();
    boneDict["Arm_L"] = new THREE.Bone();
    boneDict["Forearm_L"] = new THREE.Bone();
    boneDict["Arm_R"] = new THREE.Bone();
    boneDict["Forearm_R"] = new THREE.Bone();
    boneDict["Leg_L"] = new THREE.Bone();
    boneDict["Shin_L"] = new THREE.Bone();
    boneDict["Leg_R"] = new THREE.Bone();
    boneDict["Shin_R"] = new THREE.Bone();


    boneDict['Chest'].matrixWorld = matMul(boneDict['Spine'].matrixWorld, realBones[1].matrix);
    boneDict['Neck'].matrixWorld = matMul(boneDict['Chest'].matrixWorld, realBones[2].matrix);
    boneDict['Head'].matrixWorld = matMul(boneDict['Neck'].matrixWorld, realBones[3].matrix);
    boneDict['Arm_L'].matrixWorld = matMul(boneDict['Chest'].matrixWorld, realBones[4].matrix);
    boneDict['Forearm_L'].matrixWorld = matMul(boneDict['Arm_L'].matrixWorld, realBones[5].matrix);
    boneDict['Arm_R'].matrixWorld = matMul(boneDict['Chest'].matrixWorld, realBones[6].matrix);
    boneDict['Forearm_R'].matrixWorld = matMul(boneDict['Arm_R'].matrixWorld, realBones[7].matrix);
    boneDict['Leg_L'].matrixWorld = matMul(boneDict['Spine'].matrixWorld, realBones[8].matrix);
    boneDict['Shin_L'].matrixWorld = matMul(boneDict['Leg_L'].matrixWorld, realBones[9].matrix);
    boneDict['Leg_R'].matrixWorld = matMul(boneDict['Spine'].matrixWorld, realBones[10].matrix);
    boneDict['Shin_R'].matrixWorld = matMul(boneDict['Leg_R'].matrixWorld, realBones[11].matrix);
}

/**
* Fills the Float32Array boneArray with the bone matrices to be passed to
* the vertex shader
*/
function buildShaderBoneMatrix() {
    var c = 0;
    for (var key in boneDict) {
        for (var i = 0; i < 16; i++) {
            boneArray[c++] = boneDict[key].matrix.elements[i];
        }
    }
}

function onWindowResize() {

    camera.aspect = window.innerWidth / window.innerHeight;
    camera.updateProjectionMatrix();

    renderer.setSize(window.innerWidth, window.innerHeight);

}

function animate() {

    checkKeyboard();

    updateBody();
    requestAnimationFrame(animate);
    render();
    stats.update();

}

function render() {

    const delta = clock.getDelta();

    renderer.render(scene, camera);
}

/**
 * Retourne l'inverse de la Matrix4 M
 *
 * @param {Matrix4} M Matrice dont on cherche l'inverse
 * @return {Matrix4} L'inverse de la matrice M
 */
function getInverse(M) {
    var r = M.clone();
    return r.invert();
}

/**
* Returns a new Matrix4 as a multiplcation of m1 and m2
*
* @param {Matrix4} m1 The first matrix
* @param {Matrix4} m2 The second matrix
* @return {Matrix4} m1 x m2
*/
function matMul(m1, m2) {
    return new THREE.Matrix4().multiplyMatrices(m1, m2);
}

/**
* Returns a new Matrix4 as a translation matrix of [x,y,z]
*
* @param {number} x x component
* @param {number} y y component
* @param {number} z z component
* @return {Matrix4} The translation matrix of [x,y,z]
*/
function translation(x, y, z) {
    let mat = new THREE.Matrix4();
    mat.set(1, 0, 0, x,
        0, 1, 0, y,
        0, 0, 1, z,
        0, 0, 0, 1)
    return mat;
}

/**
* Returns a new Matrix4 as a rotation matrix of theta radians around the x-axis
*
* @param {number} theta The angle expressed in radians
* @return {Matrix4} The rotation matrix of theta rad around the x-axis
*/
function rotX(theta) {
    let c = cos(theta);
    let s = sin(theta);

    let mat = new THREE.Matrix4();
    mat.set(1, 0, 0, 0,
        0, c, -s, 0,
        0, s, c, 0,
        0, 0, 0, 1)
    return mat;
}

/**
* Returns a new Matrix4 as a rotation matrix of theta radians around the y-axis
*
* @param {number} theta The angle expressed in radians
* @return {Matrix4} The rotation matrix of theta rad around the y-axis
*/
function rotY(theta) {
    let c = cos(theta);
    let s = sin(theta);
    let mat = new THREE.Matrix4();
    mat.set(c, 0, s, 0,
        0, 1, 0, 0,
        -s, 0, c, 0,
        0, 0, 0, 1)
    return mat;
}

/**
* Returns a new Matrix4 as a rotation matrix of theta radians around the z-axis
*
* @param {number} theta The angle expressed in radians
* @return {Matrix4} The rotation matrix of theta rad around the z-axis
*/
function rotZ(theta) {
    let c = cos(theta);
    let s = sin(theta);

    let mat = new THREE.Matrix4();
    mat.set(c, -s, 0, 0,
        s, c, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1)
    return mat;
}

/**
* Returns a new Matrix4 as a scaling matrix with factors of x,y,z
*
* @param {number} x x component
* @param {number} y y component
* @param {number} z z component
* @return {Matrix4} The scaling matrix with factors of x,y,z
*/
function scale(x, y, z) {
    let mat = new THREE.Matrix4();
    mat.set(x, 0, 0, 0,
        0, y, 0, 0,
        0, 0, z, 0,
        0, 0, 0, 1)
    return mat;
}

function cos(angle) {
    return Math.cos(angle);
}

function sin(angle) {
    return Math.sin(angle);
}

function checkKeyboard() {
    for (var i = 0; i < 10; i++) {
        if (keyboard.pressed(i.toString())) {
            channel = i;
            break;
        }
    }
}

function updateBody(t) {

    switch (channel) {
        case 0:
            robot.pose_rest();
            break;
        case 1:
            robot.pose1();
            break;
        case 2:
            robot.pose2();
            break;
        case 3:
            var t = clock.getElapsedTime();
            robot.runAnimation(t);
            break;
        case 4:
            var t = clock.getElapsedTime();
            robot.customAnimation(t);
            break;
        case 5:
            break;
        case 6:
            robot.hideRobot();
            break;
        case 7:
            robot.showRobot();
            break;
        case 8:
            robot.hideHuman();
            break;
        case 9:
            robot.showHuman();
            break;
        default:
            break;
    }
}

init();
animate();
